----------------------------
--      Declaration       --
----------------------------

SwapMagic = AceLibrary("AceAddon-2.0"):new("AceEvent-2.0", "AceDB-2.0", "AceConsole-2.0", "AceHook-2.1")
local gratuity = AceLibrary("Gratuity-2.0")
local L = AceLibrary("AceLocale-2.2"):new("SwapMagic")

local _G = getfenv(0)

local displaying = false
local buttonCount = 0

L:RegisterTranslations("enUS", function() return {
	INVTYPE_GUN="Gun",
	INVTYPE_BOW="Bow",
	INVTYPE_CROSSBOW="Crossbow",
	INVTYPE_WAND="Wand",
	INVTYPE_THROWN="Thrown",
	INVTYPE_AMMO="Projectile",
	Bag = "Bag",
}end)

L:RegisterTranslations("zhCN", function() return {
	INVTYPE_GUN = "枪械",
	INVTYPE_BOW = "弩",
	INVTYPE_CROSSBOW="弩",
	INVTYPE_WAND = "魔杖",
	INVTYPE_THROWN = "投掷武器",
	INVTYPE_AMMO = "弹药",
}end)

L:RegisterTranslations("deDE", function() return {
	Bag = "Beh\195\164lter",
	INVTYPE_WAND="Zauberstab",
}end)

local slots = {
	[0] = { name=AMMOSLOT, slot="CharacterAmmoSlot"},
	[1] = { name=INVTYPE_HEAD, slot="CharacterHeadSlot"},
	[2] = { name=INVTYPE_NECK, slot="CharacterNeckSlot"},
	[3] = { name=INVTYPE_SHOULDER, slot="CharacterShoulderSlot"},
	[4] = { name=INVTYPE_BODY, slot="CharacterShirtSlot"},
	[5] = { name=INVTYPE_CHEST, slot="CharacterChestSlot"},
	[6] = { name=INVTYPE_WAIST, slot="CharacterWaistSlot"},
	[7] = { name=INVTYPE_LEGS, slot="CharacterLegsSlot"},
	[8] = { name=INVTYPE_FEET, slot="CharacterFeetSlot"},
	[9] = { name=INVTYPE_WRIST, slot="CharacterWristSlot"},
	[10]= { name=INVTYPE_HAND, slot="CharacterHandsSlot"},
	[11]= { name=INVTYPE_FINGER, slot="CharacterFinger0Slot"},
	[12]= { name=INVTYPE_FINGER, slot="CharacterFinger1Slot"},
	[13]= { name=INVTYPE_TRINKET, slot="CharacterTrinket0Slot"},
	[14]= { name=INVTYPE_TRINKET, slot="CharacterTrinket1Slot"},
	[15]= { name=INVTYPE_CLOAK, slot="CharacterBackSlot"},
	[16]= { name={INVTYPE_WEAPONMAINHAND, INVTYPE_2HWEAPON, INVTYPE_WEAPON}, slot="CharacterMainHandSlot"},
	[17]= { name={INVTYPE_WEAPONOFFHAND, INVTYPE_2HWEAPON, INVTYPE_HOLDABLE, INVTYPE_SHIELD, INVTYPE_WEAPON}, slot="CharacterSecondaryHandSlot"},
	[18]= { name={INVTYPE_RANGED, INVTYPE_RELIC, L["INVTYPE_GUN"], L["INVTYPE_BOW"], L["INVTYPE_CROSSBOW"], L["INVTYPE_WAND"], L["INVTYPE_THROWN"]}, slot="CharacterRangedSlot"},
	[19]= { name=INVTYPE_TABARD, slot="CharacterTabardSlot"}
}

----------------------------
--      Main Functions    --
----------------------------

function SwapMagic:OnInitialize()
	-- Options table
	self.options = {
		type="group",
		args = {
			quality = {
				name = "Quality", type = "text",
				desc = "Minimum quality of the items",
				get = function() return self.db.profile.MinQuality end,
				set = function(v)
					self.db.profile.MinQuality = v
				end,
				validate = {["0"] = "Poor",
							["1"] = "Common",
							["2"] = "Uncommon",
							["3"] = "Rare",
							["4"] = "Epic",
							["5"] = "Legendary",
							["6"] = "Artifact"},
				order = 2
			},
			fulltooltip = {
				name = "Full Tooltip", type = "toggle",
				desc = "Show the complete tooltip",
				get = function() return self.db.profile.FullTooltip end,
				set = function(v)
					self.db.profile.FullTooltip = v
				end,
				order = 3
			}
		}
	}
	
	-- Register everything
	self:RegisterDB("SwapMagicDB")
	self:RegisterDefaults('profile', {
		MinQuality = "1",
		FullTooltip = false,
	})
	self:RegisterChatCommand({ "/swapmagic", "/sm" }, self.options )

	if not self.version then self.version = GetAddOnMetadata("SwapMagic", "Version") end
	local rev = string.gsub(GetAddOnMetadata("SwapMagic", "X-Build"), "%$Revision: (%d+) %$", "%1")
	self.version = self.version .. " |cffff8888r"..rev.."|r"
end

function SwapMagic:OnEnable()
	self:SecureHook("PaperDollItemSlotButton_OnEnter")
	self:SecureHook("CharacterFrame_OnHide")
end

function SwapMagic:PaperDollItemSlotButton_OnEnter()
	local id = this:GetName()
	
	local i
	if IsAltKeyDown() then
		for i=0,19 do
			if slots[i].slot==id then
				id = i
				break
			end
		end

		if (tonumber(id) and ((id ~= displaying) or displaying == -1)) then
			-- no onleave function for this freaking button, so I'll have to create my own
			this:SetScript("OnLeave", function()
				self:PaperDollItemSlotButton_OnLeave()
				this.updateTooltip = nil
				GameTooltip:Hide()
				ResetCursor()
			end)
			displaying = id

			self:HideMenu()
			self:CreateMenu(id)
		end
	else
		self:HideMenu()
		displaying = -1
	end
end

function SwapMagic:CharacterFrame_OnHide()
	--self.hooks.CharacterFrame_OnHide()
	self:HideMenu()
	displaying = -1
end

function SwapMagic:PaperDollItemSlotButton_OnLeave()
	self:ScheduleEvent("SwapMagicCloseMenu", self.CloseMenu, 1, self)
end

function SwapMagic:CloseMenu()
	self:HideMenu()
	displaying = -1
end

function SwapMagic:CreateMenu(id)
	gratuity:SetInventoryItem("player", id)
	
	local items = self:GetSameTypeItemsFromBag(slots[id].name)
	local buttons = {}
	for key, item in ipairs(items) do
		table.insert(buttons, self:CreateButton(key, item, id))
	end
	
	local freeSlot = self:GetFreeSlot()
	if freeSlot then
		table.insert(buttons, self:CreateButton(table.getn(buttons)+1, freeSlot, id))
	end
	buttonCount = table.getn(buttons)
	
	for key, button in ipairs(buttons) do
		self:SetPoint(key, button, id)
	end
end

function SwapMagic:GetSameTypeItemsFromBag(types)
	local items = {}
	for bag=0,NUM_BAG_FRAMES do
		for slot=1,GetContainerNumSlots(bag) do
			gratuity:SetBagItem(bag,slot)
			if type(types) == "table" then
				for _, type in ipairs(types) do
					type = string.gsub(type, "%-", "%%%-")
					if gratuity:Find(type, 2, 4, nil, true, true) then
						local itemlink = GetContainerItemLink(bag,slot)
						local _, quality = self:GetInfoForItemLink(itemlink)
						if quality>=tonumber(self.db.profile.MinQuality) then
							table.insert(items, {["itemlink"] = itemlink, ["bag"] = bag, ["slot"] = slot})
							break
						end
					end
				end
			else
				types = string.gsub(types, "%-", "%%%-")
				if gratuity:Find(types, 2, 4, nil, true, true) then
					local itemlink = GetContainerItemLink(bag,slot)
					local _, quality = self:GetInfoForItemLink(itemlink)
					if quality>=tonumber(self.db.profile.MinQuality) then
						table.insert(items, {["itemlink"] = GetContainerItemLink(bag,slot), ["bag"] = bag, ["slot"] = slot})
					end
				end
			end
		end
	end
	
	return items
end

function SwapMagic:CreateButton(key, item, slotID)
	local button
	
	if _G["SwapMagicButton"..key] then
		button = _G["SwapMagicButton"..key]
	else
		button = CreateFrame("CheckButton", "SwapMagicButton"..key, UIParent, "ActionButtonTemplate")
		button:EnableMouse(true)
		button:SetScript("OnLeave", function() 
							self:ScheduleEvent("SwapMagicCloseMenu", self.CloseMenu, 1, self)
							GameTooltip:Hide()
						end)
		local border = button:CreateTexture("SwapMagicButton"..key.."Overlay", "OVERLAY")
		-- Credits to fizzle on the colored border!
		border:SetTexture("Interface\\Buttons\\UI-ActionButton-Border")
		border:SetVertexColor(0, 255, 0)
		border:SetBlendMode("ADD")
		border:SetAlpha(0.75)
		border:SetHeight(68)
		border:SetWidth(68)
		border:SetPoint("CENTER", button, "CENTER", 0, 1)
		border:Hide()
	end
	button:SetScript("OnEnter", function() 
						self:ShowTooltip(item) 
						self:CancelScheduledEvent("SwapMagicCloseMenu") 
					end)
	button:SetScript("OnClick", function() self:EquipItem(item, slotID) end)
	
	if item.itemlink then
		local tex, qual = self:GetInfoForItemLink(item.itemlink)
		_G["SwapMagicButton"..key.."Icon"]:SetTexture(tex)
		local r, g, b = GetItemQualityColor(qual);
		_G["SwapMagicButton"..key.."Overlay"]:SetVertexColor(r, g, b)
		_G["SwapMagicButton"..key.."Overlay"]:Show()
	else
		local slotName = this:GetName();
		local _, textureName = GetInventorySlotInfo(strsub(slotName,10));
		_G["SwapMagicButton"..key.."Icon"]:SetTexture(textureName)
		_G["SwapMagicButton"..key.."Overlay"]:Hide()
	end
	_G["SwapMagicButton"..key.."Icon"]:Show()

	button:SetFrameStrata("HIGH")
	button:Show()
	
	return button
end

function SwapMagic:GetInfoForItemLink(itemlink)
	if not itemlink then return end
	local id, itemID, itemQuality, itemName, itemTexture, itemSlot
	_,_,id = string.find(itemlink,"(item:%d+:%d+:%d+:%d+)")
	_,_,itemID = string.find(id or "","item:(%d+:%d+:%d+):%d+")
	itemName,_,itemQuality,_,_,_,_,_,itemSlot,itemTexture = GetItemInfo(id)
	
	return itemTexture, itemQuality, itemID, itemName, itemSlot
end

function SwapMagic:SetPoint(key, button, slot)
	local point, x, y = "TOPLEFT", 2, 0
	local colCount = math.ceil(math.sqrt(buttonCount))
	local col = math.fmod(key-1, colCount)
	local row = math.floor((key-1)/colCount)
	
	local relativePoint
	if slot ~= 16 and slot ~= 17 and slot ~= 18 then
		relativePoint = "TOPRIGHT"
	else
		relativePoint, x, y = "BOTTOMLEFT", 0, -2
	end
	
	if key == 1 then
		button:SetPoint(point, this, relativePoint, x, y);
	elseif math.fmod(key-1, colCount)==0 then
		button:SetPoint(point, "SwapMagicButton".. (key-colCount) , "BOTTOMLEFT", 0, -2);
	else
		button:SetPoint(point, "SwapMagicButton".. (key-1) , "TOPRIGHT", 2, 0);
	end
end

function SwapMagic:HideMenu()
	for key=0,buttonCount do
		if _G["SwapMagicButton"..key] then
			_G["SwapMagicButton"..key]:Hide()
			_G["SwapMagicButton"..key]:SetChecked(false)
		end
	end
	self:CancelScheduledEvent("SwapMagicCloseMenu")
end

-- Thanks at rabbit for the following code - taken from ClosetGnome
function SwapMagic:EquipItem(item, slot)
	local bagNum, slotNum = item.bag, item.slot
	if bagNum > -1 then
		-- GMail workaround, it hooks PickupContainerItem and watches for ALT
		local gmail
		if GMail and GMail:IsHooked("PickupContainerItem") then
			GMail:Unhook("PickupContainerItem")
			gmail = true
		end
		if not itemlink then PickupInventoryItem(displaying) end
		PickupContainerItem(bagNum, slotNum)
		if slot < 11 or slot == 15 or slot == 19 then
			AutoEquipCursorItem()
		else
			EquipCursorItem(slot)
		end
		-- Restore the hook from GMail
		if gmail then
			GMail:SecureHook("PickupContainerItem") 
		end
	end
	
	self:HideMenu()
	displaying = -1
	GameTooltip:Hide()
end

function SwapMagic:ShowTooltip(item)
	GameTooltip:SetOwner(this, "ANCHOR_RIGHT");
	GameTooltip:SetBagItem(item.bag, item.slot)
	if not self.db.profile.FullTooltip then
		local text = _G["GameTooltipTextLeft1"]:GetText()
		local textR, textG, textB = _G["GameTooltipTextLeft1"]:GetTextColor()
		GameTooltip:ClearLines()
		GameTooltip:AddLine(text, textR, textG, textB)
	end
	GameTooltip:Show()
end

function SwapMagic:IsNormalBag(bagId)
	if bagId == 0 or bagId == -1 then return true end
	local _, _, linkId = string.find(GetInventoryItemLink("player", ContainerIDToInventoryID(bagId)), "item:(%d+)")
	if not linkId then return false end
	local _, _, _, _, _, _, bagType = GetItemInfo(linkId)
	if bagType and bagType == L["Bag"] then return true end
	return false
end

-- copied from ClosetGnome
function SwapMagic:GetFreeSlot()
	for bag = 0, NUM_BAG_FRAMES, 1 do
		if self:IsNormalBag(bag) then
			local slots = GetContainerNumSlots(bag)
			for slot = 1, slots, 1 do
				local texture = GetContainerItemInfo(bag, slot)
				if not texture then
					--return bag, slot
					return {["bag"] = bag, ["slot"] = slot}
				end
			end
		end
	end
	return nil
end
